import numpy as np
import matplotlib.pyplot as plt
from urllib import request
import os
import gzip
import shutil


# Toy data related


_TOY_DATA_CLUSTER_CENTERS = ((2, -4), (-3, 3), (3, 1), (-5, -2), (-4, -5))
_TOY_DATA_COLORS = ('g', 'r', 'b', 'm', 'y')


def get_toy_data(classes=5,
                 samples_per_class=500,
                 feature_dim=2,
                 noise_dim=100,
                 feature_sigma=1.0,
                 noise_sigma=3.0,
                 cluster_centers=_TOY_DATA_CLUSTER_CENTERS,
                 test_ratio=0.1,
                 visualize=True):
    num_samples = samples_per_class * classes
    total_dim = feature_dim + noise_dim + 1
    data = np.zeros((num_samples, total_dim))
    for j in range(classes):
        # generate useful features
        data[j*samples_per_class:(j+1)*samples_per_class, 0:feature_dim] = \
            np.array(cluster_centers[j]).reshape(1, feature_dim) + \
            np.random.randn(samples_per_class, feature_dim) * feature_sigma
        # generate random features
        data[j*samples_per_class:(j+1)*samples_per_class,
             feature_dim:total_dim-1] = \
            np.random.randn(samples_per_class, noise_dim) * noise_sigma
        # generate ground-truth labels
        data[j*samples_per_class:(j+1)*samples_per_class, total_dim-1] = \
            np.ones(samples_per_class) * j
    # shuffle data
    data = np.random.permutation(data)
    # Split data into training and test set
    sp_idx = num_samples - int(test_ratio * num_samples)
    train_data = data[:sp_idx, :]
    test_data = data[sp_idx:, :]
    # visualize if asked to
    if visualize:
        plt.figure('train_data')
        train_colors = [_TOY_DATA_COLORS[int(c)] for c in train_data[:, -1]]
        plt.scatter(train_data[:, 0], train_data[:, 1], c=train_colors)
        plt.figure('test_data')
        test_colors = [_TOY_DATA_COLORS[int(c)] for c in test_data[:, -1]]
        plt.scatter(test_data[:, 0], test_data[:, 1], c=test_colors)
        plt.show()
    return train_data, test_data, classes


# MNIST data related


MNIST_CLASSES = 10
_MNIST_URL_BASE = 'http://yann.lecun.com/exdb/mnist/'
_MNIST_FILES = {'train_image': 'train-images-idx3-ubyte',
                'train_label': 'train-labels-idx1-ubyte',
                'test_image': 't10k-images-idx3-ubyte',
                'test_label': 't10k-labels-idx1-ubyte'}


def download_unzip_gz(from_url, save_as):
    print('Downloading from {}'.format(from_url))
    with request.urlopen(from_url) as src, open(save_as, 'wb') as dst:
        shutil.copyfileobj(gzip.GzipFile(fileobj=src), dst)
    print('Saved as {}'.format(save_as))


def load_MNIST_images(from_file):
    with open(from_file, 'rb') as content:
        return np.frombuffer(content.read(), np.uint8, offset=16
                             ).reshape(-1, 28*28)


def load_MNIST_labels(from_file):
    with open(from_file, 'rb') as content:
        return np.frombuffer(content.read(), np.uint8, offset=8).reshape(-1, 1)


def visualize_MNIST_data(data, rows=6, cols=10):
    out = np.zeros((28 * rows, 28 * cols))
    for r in range(rows):
        for c in range(cols):
            out[r*28:(r+1)*28, c*28:(c+1)*28] = \
                data[r+rows*c, :-1].reshape(28, 28)
    plt.figure('MNIST training image examples')
    plt.imshow(out / 255., cmap='gray')
    plt.show()


def get_MNIST_data(save_dir='.', visualize=True):
    # download MNIST data if not present already
    for filename in _MNIST_FILES.values():
        filepath = os.path.join(save_dir, filename)
        if os.path.exists(filepath):
            print('Found {}'.format(filepath))
        else:
            download_unzip_gz(_MNIST_URL_BASE+filename+'.gz', filepath)
    # load MNIST data
    train_image = load_MNIST_images(
            os.path.join(save_dir, _MNIST_FILES['train_image']))
    train_label = load_MNIST_labels(
            os.path.join(save_dir, _MNIST_FILES['train_label']))
    train_data = np.concatenate((train_image, train_label),
                                axis=1).astype(np.float64)
    test_image = load_MNIST_images(
            os.path.join(save_dir, _MNIST_FILES['test_image']))
    test_label = load_MNIST_labels(
            os.path.join(save_dir, _MNIST_FILES['test_label']))
    test_data = np.concatenate((test_image, test_label),
                               axis=1).astype(np.float64)
    # visualize examples if required
    visualize_MNIST_data(train_data)
    return train_data, test_data, MNIST_CLASSES


# optimization related


def softmax(z, dim):
    ez = np.exp(z - np.max(z, dim, keepdims=True))
    return ez / np.sum(ez, dim, keepdims=True)


def log_softmax(z, dim):
    dz = z - np.max(z, dim, keepdims=True)
    return dz - np.log(np.sum(np.exp(dz), dim, keepdims=True))


def classifier(X, W, b):
    return X @ W + b


def score(X, W, b):
    return softmax(classifier(X, W, b), 1)


def average_loss(X, y, W, b):
    y = y.astype(np.long)
    return np.mean(-log_softmax(classifier(X, W, b), 1)[np.arange(len(y)), y])


# the differentiable objective
def objective_diff(X, y, W, b, lambda1, lambda2):
    return average_loss(X, y, W, b) + 0.5 * lambda2 * np.sum(np.square(b)) + \
        0.5 * lambda1 * np.sum(np.square(W))


# the non-differentiable objective
def objective_nondiff(X, y, W, b, lambda1, lambda2):
    return average_loss(X, y, W, b) + 0.5 * lambda2 * np.sum(np.square(b)) + \
        0.5 * lambda1 * np.sum(np.sqrt(np.sum(np.square(W), 1)))


# gradient related


# TODO: gradients of the average logsoftmax loss w.r.t. W
# REMARK: not all arguments are useful, you can remove the redundant ones
def grad_W_avgloss(X, W, b, score, y):
    raise NotImplementedError


# TODO: gradients of the average logsoftmax loss w.r.t. b    # You can use e.g. scipy.optimize.minimize to compare your result for the
    # differentiable objective objective_diff.
# REMARK: not all arguments are useful, you can remove the redundant ones
def grad_b_avgloss(X, W, b, score, y):
    raise NotImplementedError


# TODO: gradients of the square regularization term for W in objective_diff
def grad_W_sqreg(W, lambda1):
    raise NotImplementedError


# TODO: gradients of the square regularization term for b in both objectives
def grad_b_sqreg(b, lambda2):
    raise NotImplementedError


if __name__ == '__main__':
    np.random.seed(43)  # manual random seed for reproducible results
    use_mnist = False  # REMARK: set to true once your code works on toy data

    # get data
    if use_mnist:
        train_data, test_data, classes = get_MNIST_data(save_dir='.',
                                                        visualize=True)
    else:
        train_data, test_data, classes = get_toy_data(visualize=True)
    X = train_data[:, :-1]
    feature_dim = X.shape[1]
    y = train_data[:, -1]
    # Normalize input data
    X = X / np.max(X)

    # initialize parameters to be trained
    W = np.random.rand(feature_dim, classes)
    b = np.zeros((1, classes))

    # TODO(optional): Do derivative check (for debugging only).
    # You can use scipy.optimize.grad_check on objective_diff to verify the 4
    # gradient functions you have implemented.
    pass

    # define optimization parameters
    lambda1, lambda2 = 0.01, 0.02  # TODO: play around to check out the effects
    max_iter = 6000
    if use_mnist:
        tau = 0.7
    else:
        tau = 1 / (np.mean(np.sum(np.square(X), axis=1)) + lambda1)
    threshold = 1e-12
    # define lists that should be filled up in the following updates
    obj_val = []
    train_error = []
    test_error = []
    # perform proximal gradient iterations
    for i in range(max_iter):

        # TODO: implement proximal gradient here...
        raise NotImplementedError

        # evalualte stopping criterion
        if i > 0 and obj_val[i-1] - obj_val[i] < threshold:
            break

    # Plot objective value, train / test loss
    fig = plt.figure('results')
    ax = fig.add_subplot(1, 3, 1)
    ax.plot(range(len(obj_val)), obj_val, 'r')
    ax = fig.add_subplot(1, 3, 2)
    ax.plot(range(len(train_error)), train_error, 'g')
    ax = fig.add_subplot(1, 3, 3)
    ax.plot(range(len(test_error)), test_error, 'b')
    plt.show()
    # plot classifier for toy data case
    if not use_mnist:
        s = 0.1
        x1Grid, x2Grid = np.meshgrid(np.arange(-4, 4, s), np.arange(-4, 4, s))
        xGrid = np.stack((x1Grid.flatten(), x2Grid.flatten()), axis=1)
        scores_train = classifier(xGrid, W[0:2, :], b)
        plt.figure('classifier')
        test_colors = [_TOY_DATA_COLORS[int(c)] for c in test_data[:, -1]]
        plt.scatter(test_data[:, 0], test_data[:, 1], c=test_colors)
        for c in range(classes):
            color = _TOY_DATA_COLORS[c]
            plt.contour(x1Grid, x2Grid,
                        scores_train[:, c].reshape(x1Grid.shape),
                        levels=1, colors=color)
        plt.show()
